#######################################################################
# Functions to fit multivariate frailty models for Rose data using
# 1. MCEM
# 2. Penalized partial likelihood
#######################################################################

# Function to create design matrices:

design.z <- function(x)
{
    q <- length(unique(x))
    n <- length(x)
    z <- matrix(0,ncol=q, nrow=length(x))
    key <- cbind(sort(unique(x)), 1:q)
    col1 <- match(x, key[,1])
    for (i in 1:n) z[i,col1[i]] <- 1
    z
}

# Modified verision of the control function for penalized partial likelihood estimation

frailty.controlgauss <- function (opt, iter, old, fcoef, trH, loglik) 
{
    if (iter == 0) {
        if (!is.null(opt$theta)) 
            theta <- opt$theta
        else {
            if (is.null(opt$init)) 
                theta <- sam.theta.start
            else theta <- opt$init[1]
        }
        list(theta = theta)
    }
    else {
        if (is.null(opt$trace)) 
            trace <- F
        else trace <- opt$trace
        nfrail <- length(fcoef)
        fsum <- sum(fcoef^2)
        theta <- old$theta
        resid <- fsum/(nfrail - trH/theta) - theta
        if (iter == 1) {
            history <- c(theta = theta, resid = resid, fsum = fsum, 
                trace = trH)
            if (is.null(opt$init)) {
                if (resid > 0) 
                  theta <- theta * 2
                else theta <- theta/2
            }
            else theta <- opt$init[2]
            list(theta = theta, done = F, history = history)
        }
        else {
            history <- rbind(old$history, as.vector(c(theta, 
                resid, fsum, trH)))
            if (iter == 2) {
                if (all(history[, 2] > 0)) 
                  theta <- history[2, 1] * 1.5
                else if (all(history[, 2] < 0)) 
                  theta <- history[2, 1]/1.5
                else theta <- mean(history[1:2, 1])
                if (trace) {
                  print(history)
                  cat("    new theta=", theta, "\n\n")
                }
                list(theta = theta, done = F, history = history)
            }
            else {
                done <- (abs(history[iter, 2]) < opt$eps)
                newtheta <- theta + sam.mult*resid
                if(newtheta<0.0000001) newtheta <- 0.0000001
                if (trace) {
                  print(history)
                  cat("    new theta=", format(newtheta), "\n\n")
                }
                list(theta = newtheta, done = done, history = history)
            }
        }
    }
}

# Functions for sampling in the E-step using rejection sampling:
# - Shared frailty model

RS.par <- function(stat, x, z, c.haz, beta, theta, b.start, cc, itermax)
{
  b.current <- b.start
  q <- dim(z)[2]
  iter <- 1
  b.diff <- 100
  while ((iter < itermax) && (sum(b.diff^2) > cc))
    {
      lp <- x %*% beta + z %*% b.current
      haz.lp <- c.haz * exp(lp)
      b.1 <- t(z) %*% (stat - haz.lp) - b.current/theta
      b.2 <- diag(c(1/(t(z) %*% haz.lp + 1/theta)))
      b.diff <- b.2 %*% b.1
      b.current <- b.current + b.diff
      iter <- iter + 1
      cat(sum(b.diff^2), "\n")
    }
  lp <- x %*% beta + z %*% b.current
  haz.lp <- c.haz * exp(lp)
  f.max1 <- sum(ifelse(stat,  lp -  haz.lp, - haz.lp))
  f.max2 <- -(0.5/theta) * sum(b.current^2)
  f.max <- f.max1 + f.max2
  return(list(f.max=f.max, b.new=b.current, b.var=b.2, iter=iter))
}

RS.sample <- function(b.sample, nosample, q, theta, c.haz, beta, x, z, stat, tmp.pars, var.infl)
  {
    w <- runif(nosample)
    accept <- rep(NA, nosample)
    f.prop <- rep(NA, nosample)
    f.samp <- rep(NA, nosample)
    ratio <- rep(NA, nosample)
    imat <- solve(tmp.pars$b.var)
    det.V <- prod(eigen(tmp.pars$b.var, only.values=T)$values)
    for (i in 1:nosample)
      {
        lp <- x %*% beta + z %*% b.sample[,i]
        haz.lp <- c.haz * exp(lp)
        f.prop[i] <- sum(ifelse(stat, lp  -  haz.lp, - haz.lp)) - (0.5/theta) * sum(b.sample[,i]^2) 
        f.samp[i] <- -0.5* t(b.sample[,i] - tmp.pars$b.new) %*% imat %*% (b.sample[,i] - tmp.pars$b.new)  
      }
    max.diff <- max(f.prop-f.samp)
    ratio <- exp(f.prop - f.samp - max.diff)
    accept <- w < ratio
    accept.ratio <- sum(accept)/nosample
    acc.b.sample <- b.sample[,accept]
    return(acc.b.sample, b.sample, accept, accept.ratio, w, ratio, f.prop, f.samp)
  }

# Functions for sampling in the E-step using rejection sampling:
# - 4 variance component frailty model

RS.par4 <- function(stat, x, z, c.haz, beta, theta, dim.theta, b.start, cc, itermax)
{
  b.current <- b.start
  q <- dim(z)[2]
  iter <- 1
  b.diff <- 100
  while ((iter < itermax) && (sum(b.diff^2) > cc))
    {
      lp <- x %*% beta + z %*% b.current
      haz.lp <- c.haz * exp(lp)
      b.1 <- t(z) %*% (stat - haz.lp) -
             b.current/c(rep(theta[1],dim.theta[1]),rep(theta[2],dim.theta[2]),rep(theta[3],dim.theta[3]),rep(theta[4],dim.theta[4]))
      b.2 <- diag(c(1/(t(z) %*% haz.lp +
          1/c(rep(theta[1],dim.theta[1]),rep(theta[2],dim.theta[2]),rep(theta[3],dim.theta[3]),rep(theta[4],dim.theta[4])))))
      b.diff <- b.2 %*% b.1
      b.current <- b.current + b.diff
      iter <- iter + 1
      cat(sum(b.diff^2), "\n")
    }
  lp <- x %*% beta + z %*% b.current
  haz.lp <- c.haz * exp(lp)
  f.max <- 0
  return(list(f.max=f.max, b.new=b.current, b.var=b.2, iter=iter))
}

RS.sample4 <- function(b.sample, nosample, q, theta.mat, c.haz, beta, x, z, stat, tmp.pars, var.infl)
  {
    w <- runif(nosample)
    accept <- rep(NA, nosample)
    f.prop <- rep(NA, nosample)
    f.samp <- rep(NA, nosample)
    ratio <- rep(NA, nosample)
    imat <- solve(tmp.pars$b.var)
    imat.theta <- solve(theta.mat)
    det.V <- prod(eigen(tmp.pars$b.var, only.values=T)$values)
    for (i in 1:nosample)
      {
        lp <- x %*% beta + z %*% b.sample[,i]
        haz.lp <- c.haz * exp(lp)
        f.prop[i] <- sum(ifelse(stat, lp  -  haz.lp, - haz.lp)) - (0.5) * (t(b.sample[,i]) %*% imat.theta %*% b.sample[,i])
        f.samp[i] <- -0.5* t(b.sample[,i] - tmp.pars$b.new) %*% imat %*% (b.sample[,i] - tmp.pars$b.new)
      }
    max.diff <- max(f.prop-f.samp)
    ratio <- exp(f.prop - f.samp - max.diff)
    accept <- w < ratio
    accept.ratio <- sum(accept)/nosample
    acc.b.sample <- b.sample[,accept]
    return(acc.b.sample, b.sample, accept, accept.ratio, w, ratio, f.prop, f.samp)
  }

# Read the data in

r <- read.table("r2.dat", header=T, sep="\t")
taimi2 <- r$ruutu * 10 + r$taimi

taimi.key <- data.frame(cbind(id=unique(taimi2),id2=1:length(unique(taimi2))))
t2 <- cbind(taimi2, id2=match(taimi2, taimi.key[,1]))

taimi.z <- design.z(t2[,2])

ruutu.key <- data.frame(cbind(id=unique(r$ruutu),id2=1:length(unique(r$ruutu))))
ruutu2 <- cbind(r$ruutu, id2=match(r$ruutu, ruutu.key[,1]))

ruutu.z <- design.z(ruutu2[,2])

lohko.key <- data.frame(cbind(id=unique(r$lohko),id2=1:length(unique(r$lohko))))
lohko2 <- cbind(r$lohko, id2=match(r$lohko, lohko.key[,1]))

lohko.z <- design.z(lohko2[,2])

z <- cbind(lohko.z, ruutu.z, taimi.z)
time <- r$kesto
stat <- r$pois.syy==2
x <- design.z(r$kasittely)[,-1]
x <- cbind(design.z(r$kasittely)[,-3],design.z(r$lohko)[,-4])

# Add a random number from N(0,0.1) to times
time2 <- time + rnorm(length(time),0,0.1)
x <- design.z(r$kasittely)[,-3]
x.lohko <- design.z(r$lohko)[,-4]

# set parameter value for frailty.controlgauss() 

sam.mult <- 3

# Shared frailty model using penalized partial likelihood

f3b <- coxph(Surv(time2,stat)~x+x.lohko+frailty.gaussian(t2[,2], sparse=T))

# Standard error for theta
fvar <- f3b$fvar
K.sq <- sum(fvar^2)
K.tr <- sum(fvar)
f3b.theta <- f3b$history$"frailty.gaussian(t2[, 2], sparse = T)"$theta
f3b.se.theta <- sqrt(2*(f3b.theta^2)/(224+(1/f3b.theta^2)*K.sq-(2/f3b.theta)*K.tr))

# Frailty model with 4 var comp
f4b <- coxph(Surv(time2,stat)~x+x.lohko+frailty.gaussian(x.full[,1]*t2[,2], sparse=F)+frailty.gaussian(x.full[,2]*t2[,2], sparse=F)+
    frailty.gaussian(x.full[,3]*t2[,2], sparse=F)+frailty.gaussian(x.full[,4]*t2[,2], sparse=F), outer.max=30)

# Standard errors for thetas
f4b.frail <- (coef(f4b)[names(coef(f4b))!="gamma:0"])[7:230]
f4b.fvar <- diag(f4b$var[c(8:63,65:123,125:177,179:234),c(8:63,65:123,125:177,179:234)])
K.sq <- c(sum(f4b.fvar[1:56]^2),sum(f4b.fvar[57:115]^2),sum(f4b.fvar[116:168]^2),sum(f4b.fvar[169:224]^2))
K.tr <- c(sum(f4b.fvar[1:56]),sum(f4b.fvar[57:115]),sum(f4b.fvar[116:168]),sum(f4b.fvar[169:224]))
f4b.theta <- c(f4b$history$"frailty.gaussian(x.full[, 1] * t2[, 2], sparse = F)"$theta,
               f4b$history$"frailty.gaussian(x.full[, 2] * t2[, 2], sparse = F)"$theta,
               f4b$history$"frailty.gaussian(x.full[, 3] * t2[, 2], sparse = F)"$theta,
               f4b$history$"frailty.gaussian(x.full[, 4] * t2[, 2], sparse = F)"$theta)
f4b.se.theta <- c(sqrt(2*(f4b.theta[1]^2)/(56+(1/f4b.theta[1]^2)*K.sq[1]-(2/f4b.theta[1])*K.tr[1])),
                  sqrt(2*(f4b.theta[2]^2)/(59+(1/f4b.theta[2]^2)*K.sq[2]-(2/f4b.theta[2])*K.tr[2])),
                  sqrt(2*(f4b.theta[3]^2)/(53+(1/f4b.theta[3]^2)*K.sq[3]-(2/f4b.theta[3])*K.tr[3])),
                  sqrt(2*(f4b.theta[4]^2)/(56+(1/f4b.theta[4]^2)*K.sq[4]-(2/f4b.theta[4])*K.tr[4])))

################################################################################
# MCEM for shared frailty model:

ord <- order(time2)
time.s <- time2[ord]
stat.s <- stat[ord]==1
x <- cbind(design.z(r$kasittely)[,-3],design.z(r$lohko)[,-4])
x.s <- x[ord,]
z.s <- taimi.z[ord,]

# Initial values from the penalized partial likelihood estimation
beta <- coef(f3b)[1:6]; theta <- 0.2196

e.lpred <- exp(x.s %*% beta + z.s %*% f3b$frail)
tmp.chaz <- cum.haz(e.lpred, time.s, stat.s)$chaz
b.ave <- f3b$frail; inum <- 224
total.sample <- 100; nosample <- 1000
conv.flag <- T; c.limit <- 0.006; max.iter <- 100; k <- 1

date()
betatheta <- matrix(c(beta,theta,NA,NA,100), ncol=10, nrow=1);
while ( k < max.iter && conv.flag)
  {
    tmp.pars <- RS.par(stat.s, x.s, z.s, tmp.chaz, beta, theta, b.ave, 0.00001, 25)
    v.trial <- diag(tmp.pars$b.var)
    s.num <- 0; a.sample <- rep(0,length(b.ave)); j <- 0
    while (s.num < total.sample)
      {
        j <- j + 1
        b.sample <- matrix(rnorm(inum*nosample, tmp.pars$b.new, sqrt(v.trial)), nrow=inum)
        tmp.sample <- RS.sample(b.sample, nosample, inum, theta, tmp.chaz, beta, x.s, z.s, stat.s, tmp.pars, var.infl)
        write.table(tmp.sample$acc.b.sample, paste("Sample_",j,sep=""),quote=F,row.names=F,col.names=F)
        s.num <- s.num + sum(tmp.sample$accept)
        cat(j,": Acceptance ratio:", tmp.sample$accept.ratio, "# of samples:", s.num, "\n")
      }
    p <- dim(x.s)[2];  q <- dim(z.s)[2];    n <- length(time.s)
    e.za <- matrix(NA,n,j); m.j <- rep(NA,j)
    b.sq <- matrix(NA,q,j); b.ave2 <- matrix(NA,q,j)
    for (i in 1:j) {
        b.sample <- as.matrix(read.table(paste("Sample_",i,sep=""),sep=" "))
        e.za[,i] <- apply(exp(z.s %*% b.sample), 1, sum)
        b.sq[,i] <- apply(b.sample^2,1,sum)
        b.ave2[,i] <- apply(b.sample,1,sum)
        m.j[i] <- dim(b.sample)[2]
        }
    zb.ave <- log(apply(e.za,1,sum)/sum(m.j))
    theta.new <- mean(apply(b.sq,1,sum)/sum(m.j))
    b.ave <- apply(b.ave2,1,sum)/sum(m.j)
    tmp.fit <- coxph(Surv(time.s,stat.s)~x.s+offset(zb.ave))
    beta.new <- coef(tmp.fit)
    var.beta <- tmp.fit$var
    chaz.new <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$chaz
    beta.diff <- max(abs(beta-beta.new)/(abs(beta)+0.001)); theta.diff <- max(abs(theta-theta.new)/(abs(theta)+0.001))
    diff <- max(c(beta.diff, theta.diff))
    cat(k, ":", beta.new, theta.new, diff, total.sample,"\n")
    beta <- beta.new; theta <- theta.new
    betatheta <- rbind(betatheta,c(beta.new, theta.new, diff, NA, total.sample))
    tmp.chaz <- chaz.new
    if (k>2)
        {
        betatheta[k+1,p+3] <- var(betatheta[(k-1):(k+1),p+2])/mean(betatheta[(k-1):(k+1),p+2])
        cat("CV:", betatheta[k+1,p+3], "\n")
        if (k>30 && betatheta[k+1,p+3]>betatheta[k,p+3]) total.sample <- total.sample+0.5*total.sample
        if (k>30 && max(betatheta[(k-1):(k+1),p+2]) < c.limit) conv.flag <- F
        }
    k <- k + 1
    if (total.sample > nosample && nosample < 3000) nosample <- 1.5 * nosample
  }
date()

betatheta.roseshared <- betatheta

# Variance using Louis's (1982) formula
# Squared score

theta.new <- betatheta.roseshared[60,7]
beta.new <- betatheta.roseshared[60,1:6]

apu <- sample(1:300,10,FALSE,rep(1/300,300))
tmp.sample <- as.matrix(cbind(read.table(paste("Sample_",apu[1],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[2],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[3],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[4],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[5],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[6],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[7],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[8],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[9],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[10],sep=""),sep=" ")))
tmp.m <- dim(tmp.sample)[2]
za <- z.s %*% tmp.sample
zb.ave <- log(apply(exp(z.s %*% tmp.sample), 1, mean))
mart.res <- matrix(0, nrow=n, ncol=tmp.m)
chaz.new <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$chaz
for(i in 1:tmp.m)  mart.res[,i] <- stat.s - chaz.new * exp(x.s %*% beta.new + za[,i])
sco.beta <- t(x.s) %*% mart.res
eka.theta <- q/theta.new
toka.theta <- apply(tmp.sample^2,2,sum)/theta.new^2
sco.theta <- -0.5 * (eka.theta - toka.theta)
tmp.haz <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$hazard[stat.s]
d <- length(time.s[stat.s])
tmp.rw <- matrix(NA, nrow=d, ncol=tmp.m)
atrisk <- t( sapply(time.s, function(a,b){ rep(a, length(b)) >= b }, b=time.s[stat.s]))
xbeta <- x.s %*% beta.new
for ( i in 1:tmp.m) {
  rr <- exp(xbeta + za[,i])
  tmp.rw[,i] <- apply( atrisk * matrix(rep(rr,d), ncol=d), 2, sum)
 }
sco.lambda <- matrix(rep(1/tmp.haz, tmp.m), ncol=tmp.m) - tmp.rw
score.btl <- rbind(sco.beta,c(sco.theta),sco.lambda)
score.sq <- score.btl %*% t(score.btl)/tmp.m

# Estimated variance matrix
tmp.beta <- array(0,c(p,p,tmp.m))
tmp.bl <- array(0,c(d,p,tmp.m))
date()
for (i in 1:tmp.m)
{
    rr <- exp(xbeta + za[,i])
    chaz.lp <- (chaz.new * rr)
    for (j in 1:n) tmp.beta[,,i] <- tmp.beta[,,i] + x.s[j,] %*% t(x.s[j,]) * chaz.lp[j]
    tmp.bl[,,i] <- t(atrisk) %*% (x.s * c(rr))
}
I.beta <- apply(tmp.beta,c(1,2),mean)
date()
I.bl <- apply(tmp.bl,c(1,2),mean)

I.lambda <- diag(1/tmp.haz^2)
I.theta <- -0.5 * (q/theta.new^2 - 2 * mean(apply(tmp.sample^2,2,sum))/theta.new^3)
I.btl <- cbind(rbind(I.beta,rep(0,6),I.bl),
               rbind(matrix(rep(0,p), ncol=1),I.theta,matrix(rep(0,d), ncol=1)),
               rbind(t(I.bl), rep(0,d), I.lambda))
I.obs <- I.btl - score.sq
var.btl <- solve(I.obs)
sqrt(diag(var.btl[1:7,1:7]))

# Standard error estimation may be repeated several times and the mean can be taken over the replications!

######################################################################################
# Larger model with 4 variance components

beta <- coef(f4b)[1:6]; theta <- c(0.299,0.268,0.207,0.121)
dim.theta <- c(56,59,53,56)
cdim.theta <- cumsum(dim.theta)
index.theta <- c(rep(1,dim.theta[1]),rep(2,dim.theta[2]),rep(3,dim.theta[3]),rep(4,dim.theta[4]))
theta.mat <- diag(c(rep(theta[1],dim.theta[1]),rep(theta[2],dim.theta[2]),rep(theta[3],dim.theta[3]),rep(theta[4],dim.theta[4])))

e.lpred <- exp(x.s %*% beta + z.s %*% f4b.frail)
tmp.chaz <- cum.haz(e.lpred, time.s, stat.s)$chaz
b.ave <- f4b.frail; inum <- 224
total.sample <- 100; nosample <- 1000
conv.flag <- T; c.limit <- 0.006; max.iter <- 100; k <- 1

date()
betatheta <- matrix(c(beta,theta,NA,NA,100), ncol=13, nrow=1);
while ( k < max.iter && conv.flag)
  {
    tmp.pars <- RS.par4(stat.s, x.s, z.s, tmp.chaz, beta, theta, dim.theta, b.ave, 0.00001, 25)
    v.trial <- diag(tmp.pars$b.var)
    s.num <- 0; a.sample <- rep(0,length(b.ave)); j <- 0
    while (s.num < total.sample)
      {
        j <- j + 1
        b.sample <- matrix(rnorm(inum*nosample, tmp.pars$b.new, sqrt(v.trial)), nrow=inum)
        tmp.sample <- RS.sample4(b.sample, nosample, inum, theta.mat, tmp.chaz, beta, x.s, z.s, stat.s, tmp.pars, var.infl)
        write.table(tmp.sample$acc.b.sample, paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",j,sep=""),quote=F,row.names=F,col.names=F)
        s.num <- s.num + sum(tmp.sample$accept)
        cat(j,": Acceptance ratio:", tmp.sample$accept.ratio, "# of samples:", s.num, "\n")
      }
    p <- dim(x.s)[2];  q <- dim(z.s)[2];    n <- length(time.s)
    e.za <- matrix(NA,n,j); m.j <- rep(NA,j)
    b.sq <- matrix(NA,q,j); b.ave2 <- matrix(NA,q,j)
    for (i in 1:j) {
        b.sample <- as.matrix(read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",i,sep=""),sep=" "))
        e.za[,i] <- apply(exp(z.s %*% b.sample), 1, sum)
        b.sq[1:cdim.theta[1],i]                 <- apply(b.sample[1:cdim.theta[1],]^2,1,sum)
        b.sq[(cdim.theta[1]+1):cdim.theta[2],i] <- apply(b.sample[(cdim.theta[1]+1):cdim.theta[2],]^2,1,sum)
        b.sq[(cdim.theta[2]+1):cdim.theta[3],i] <- apply(b.sample[(cdim.theta[2]+1):cdim.theta[3],]^2,1,sum)
        b.sq[(cdim.theta[3]+1):cdim.theta[4],i] <- apply(b.sample[(cdim.theta[3]+1):cdim.theta[4],]^2,1,sum)
        b.ave2[,i] <- apply(b.sample,1,sum)
        m.j[i] <- dim(b.sample)[2]
        }
    zb.ave <- log(apply(e.za,1,sum)/sum(m.j))
    theta.new <- tapply(apply(b.sq,1,sum)/sum(m.j),index.theta,mean)
    theta.mat <- diag(c(rep(theta.new[1],dim.theta[1]),rep(theta.new[2],dim.theta[2]),
                        rep(theta.new[3],dim.theta[3]),rep(theta.new[4],dim.theta[4])))
    b.ave <- apply(b.ave2,1,sum)/sum(m.j)
    tmp.fit <- coxph(Surv(time.s,stat.s)~x.s+offset(zb.ave))
    beta.new <- coef(tmp.fit)
    var.beta <- tmp.fit$var
    chaz.new <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$chaz
    beta.diff <- max(abs(beta-beta.new)/(abs(beta)+0.001)); theta.diff <- max(abs(theta-theta.new)/(abs(theta)+0.001))
    diff <- max(c(beta.diff, theta.diff))
    cat(k, ":", beta.new, theta.new, diff, total.sample,"\n")
    beta <- beta.new; theta <- theta.new
    betatheta <- rbind(betatheta,c(beta.new, theta.new, diff, NA, total.sample))
    tmp.chaz <- chaz.new
    if (k>2)
        {
        betatheta[k+1,p+6] <- var(betatheta[(k-1):(k+1),p+5])/mean(betatheta[(k-1):(k+1),p+5])
        cat("CV:", betatheta[k+1,p+6], "\n")
        if (k>30 && betatheta[k+1,p+6]>betatheta[k,p+6]) total.sample <- total.sample+0.5*total.sample
        if (k>30 && max(betatheta[(k-1):(k+1),p+5]) < c.limit) conv.flag <- F
        }
    k <- k + 1
    if (total.sample > nosample && nosample < 3000) nosample <- 1.5 * nosample
  }
date()

betatheta.rose4 <- betatheta

# Variance using Louis's (1982) formula
# Squared score

theta.new <- betatheta.rose4[59,7:10]
beta.new <- betatheta.rose4[59,1:6]

apu <- sample(1:328,10,FALSE,rep(1/328,328))
tmp.sample <- as.matrix(cbind(read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[1],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[2],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[3],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[4],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[5],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[6],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[7],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[8],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[9],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[10],sep=""),sep=" ")))
tmp.m <- dim(tmp.sample)[2]
za <- z.s %*% tmp.sample
zb.ave <- log(apply(exp(z.s %*% tmp.sample), 1, mean))
mart.res <- matrix(0, nrow=n, ncol=tmp.m)
chaz.new <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$chaz
for(i in 1:tmp.m)  mart.res[,i] <- stat.s - chaz.new * exp(x.s %*% beta.new + za[,i])
sco.beta <- t(x.s) %*% mart.res
tmp.eka <- dim.theta/theta.new
eka.theta <- rbind(rep(tmp.eka[1],tmp.m),rep(tmp.eka[2],tmp.m),rep(tmp.eka[3],tmp.m),rep(tmp.eka[4],tmp.m))
toka.theta <- rbind(apply(tmp.sample[1:cdim.theta[1],]^2,2,sum)/(theta.new[1]^2),
                apply(tmp.sample[(cdim.theta[1]+1):cdim.theta[2],]^2,2,sum)/(theta.new[2]^2),
                apply(tmp.sample[(cdim.theta[2]+1):cdim.theta[3],]^2,2,sum)/(theta.new[3]^2),
                apply(tmp.sample[(cdim.theta[3]+1):cdim.theta[4],]^2,2,sum)/(theta.new[4]^2))
sco.theta <- -0.5 * (eka.theta - toka.theta)
tmp.haz <- cum.haz(exp(x.s %*% beta.new + zb.ave), time.s, stat.s)$hazard[stat.s]
d <- length(time.s[stat.s])
tmp.rw <- matrix(NA, nrow=d, ncol=tmp.m)
atrisk <- t( sapply(time.s, function(a,b){ rep(a, length(b)) >= b }, b=time.s[stat.s]))
xbeta <- x.s %*% beta.new
tmp.beta <- array(0,c(p,p,tmp.m))
tmp.bl <- array(0,c(d,p,tmp.m))
date()
for ( i in 1:tmp.m) {
  rr <- exp(xbeta + za[,i])
  tmp.rw[,i] <- apply( atrisk * matrix(rep(rr,d), ncol=d), 2, sum)
  chaz.lp <- (chaz.new * rr)
  for (j in 1:n) tmp.beta[,,i] <- tmp.beta[,,i] + x.s[j,] %*% t(x.s[j,]) * chaz.lp[j]
  tmp.bl[,,i] <- t(atrisk) %*% (x.s * c(rr))
 }
date()
sco.lambda <- matrix(rep(1/tmp.haz, tmp.m), ncol=tmp.m) - tmp.rw
score.btl <- rbind(sco.beta,sco.theta,sco.lambda)
score.sq <- score.btl %*% t(score.btl)/tmp.m

# Estimated variance matrix
I.beta <- apply(tmp.beta,c(1,2),mean)
I.bl <- apply(tmp.bl,c(1,2),mean)

I.lambda <- diag(1/tmp.haz^2)
I.theta <- diag(-0.5 * (dim.theta/(theta.new^2) - (2 *
           c(mean(apply(tmp.sample[1:cdim.theta[1],]^2,2,sum))/(theta.new[1]^3),
           mean(apply(tmp.sample[(cdim.theta[1]+1):cdim.theta[2],]^2,2,sum))/(theta.new[2]^3),
           mean(apply(tmp.sample[(cdim.theta[2]+1):cdim.theta[3],]^2,2,sum))/(theta.new[3]^3),
           mean(apply(tmp.sample[(cdim.theta[3]+1):cdim.theta[4],]^2,2,sum))/(theta.new[4]^3)))))
I.btl <- cbind(rbind(I.beta,matrix(rep(0,6*4), ncol=6),I.bl),
               rbind(matrix(rep(0,p*4), ncol=4),I.theta,matrix(rep(0,d*4), ncol=4)),
               rbind(t(I.bl), matrix(rep(0,d*4),ncol=d), I.lambda))
I.obs <- I.btl - score.sq
var.btl <- solve(I.obs)
sqrt(diag(var.btl[1:10,1:10]))

# Standard error estimation may be repeated several times and the mean can be taken over the replications!

######################################################################################
# LR-test for the two models

theta.new <- betatheta.roseshared[60,7]
beta.new <- betatheta.roseshared[60,1:6]

tmp.logl <- rep(NA, 30)
for (j in 1:30) {
# 10 bulks of samples
apu <- sample(1:300,10,FALSE,rep(1/300,300))
idx.simple <- apu
tmp.sample <- as.matrix(cbind(read.table(paste("Sample_",apu[1],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[2],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[3],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[4],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[5],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[6],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[7],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[8],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[9],sep=""),sep=" "),
                              read.table(paste("Sample_",apu[10],sep=""),sep=" ")))

tmp.m <- dim(tmp.sample)[2]
za <- z.s %*% tmp.sample
lzb.ave <- rep(NA, 716); zb.ave <- rep(NA, 716)
for(i in 1:716) {lzb.ave[i] <- log(mean(exp(za[i,])))
                 zb.ave[i]  <- mean(za[i,])}
bsq.ave <- mean(apply(tmp.sample^2,2,sum))
tmp.haz <- cum.haz(exp(x.s %*% beta.new + lzb.ave), time.s, stat.s)
chaz.new <- tmp.haz$chaz
haz.new <- tmp.haz$hazard

pen.shared <- -0.5*(224*log(theta.new)+bsq.ave/theta.new)
logl.shared <- sum(log(haz.new^stat.s)+stat.s*(x.s %*% beta.new + zb.ave)-chaz.new*exp(x.s %*% beta.new + lzb.ave))+pen.shared
tmp.logl[j] <- logl.shared
}

logl.shared <- mean(tmp.logl)

# The # of samples is enough, only small fluctuations in tmp.logl

# Larger model

theta.new <- betatheta.rose4[59,7:10]
beta.new <- betatheta.rose4[59,1:6]

tmp.logl2 <- rep(NA, 30)
for (j in 1:30) {
apu <- sample(1:328,10,FALSE,rep(1/328,328))
idx.4 <- apu
tmp.sample <- as.matrix(cbind(read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[1],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[2],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[3],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[4],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[5],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[6],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[7],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[8],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[9],sep=""),sep=" "),
                              read.table(paste("C:\\My Documents\\MCEM\\Rose_model2\\Sample_",apu[10],sep=""),sep=" ")))
tmp.m <- dim(tmp.sample)[2]

za <- z.s %*% tmp.sample
lzb.ave <- rep(NA, 716); zb.ave <- rep(NA, 716)
for(i in 1:716) {lzb.ave[i] <- log(mean(exp(za[i,])))
                 zb.ave[i]  <- mean(za[i,])}

bsq.ave <- c(mean(apply(tmp.sample[1:cdim.theta[1],]^2,2,sum)),
           mean(apply(tmp.sample[(cdim.theta[1]+1):cdim.theta[2],]^2,2,sum)),
           mean(apply(tmp.sample[(cdim.theta[2]+1):cdim.theta[3],]^2,2,sum)),
           mean(apply(tmp.sample[(cdim.theta[3]+1):cdim.theta[4],]^2,2,sum)))

tmp.haz <- cum.haz(exp(x.s %*% beta.new + lzb.ave), time.s, stat.s)
chaz.new <- tmp.haz$chaz
haz.new <- tmp.haz$hazard

pen.4 <- -0.5*sum(dim.theta*log(theta.new)+bsq.ave/theta.new)
logl.4 <- sum(log(haz.new^stat.s)+stat.s*(x.s %*% beta.new + zb.ave)-chaz.new*exp(x.s %*% beta.new + lzb.ave))+pen.4
tmp.logl2[j] <- logl.4
}

logl.4 <- mean(tmp.logl2)

LR.rose <- 2*(logl.4-logl.shared)
1-pchisq(LR.rose, 3)
